# coding: utf-8

# region License

# Copyright (c) 2018 Alexandre Vaissière <avaiss@fmiw.org>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# endregion

"""
Parses key URI, as specified in
https://github.com/google/google-authenticator/wiki/Key-Uri-Format.
"""

import logging
import typing

from urllib.parse import urlsplit, parse_qs, unquote

log = logging.getLogger(__name__)

_DEFAULTS = {
    'algorithm': 'SHA1',
    'digits': 6,
    'period': 30,
}


def normalize_name(label: str, issuer: str=None) -> typing.Tuple[str, typing.Optional[str]]:
    """Ensures there is no redundancy between label and issuer, as labels generally are prefixed with issuer
    and a colon.

    >>> normalize_name('Amazon:ham@spam.org', 'Amazon')
    ('ham@spam.org', 'Amazon')
    >>> normalize_name('Amazon:ham@spam.org')
    ('ham@spam.org', 'Amazon')
    >>> normalize_name('ham@spam.org')
    ('ham@spam.org', None)
    """
    if not issuer:
        head, sep, tail = label.partition(':')
        if tail:
            return tail, head
        return head, None

    if label.startswith(issuer + ':'):
        label = label[len(issuer) + 1:]
    return label, issuer


def parse_uri(url: str) -> dict:
    """Parses the given otpauth URI into a dictionary. The resulting dictionary will have the
    following entries:

    - oath_type (str): totp or hotp (mandatory),
    - label (str): the label, generally an account name (mandatory),
    - secret (str): the key, as a b32 encoding string (mandatory),
    - issuer (str): the issuer, generally the website name,
    - algorithm (str): the hash algorithm to use (mandatory),
    - digits (int): the number of digits the resulting code must keep (mandatory),
    - counter (int): the initial counter for hotp credentials (mandatory for hotp),
    - periodr (int): the period for totp credentials (mandatory for totp),

    :param url: the uri to parse
    :return: all parameters as a dictionary.
    """
    split = urlsplit(url)
    log.info(split)
    query = parse_qs(split.query)

    label = unquote(split.path[1:])
    issuer = query.get('issuer', [None])[0]

    label, issuer = normalize_name(label, issuer)

    res = {
        'oath_type': split.netloc,
        'label': label,
        'secret': query['secret'][0],
        'issuer': issuer,
    }

    for optional in ('algorithm', 'digits', 'counter', 'period'):
        if optional in query:
            res[optional] = query[optional][0]
        elif optional in _DEFAULTS:
            res[optional] = _DEFAULTS[optional]

    log.debug('Parsed %s into %s', url, res)

    return res



# coding=utf-8

# region License

# Copyright (c) 2017,2018 Alexandre Vaissière <avaiss@fmiw.org>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# endregion

import logging
import tempfile

from gi.repository import Gdk
from gi.repository import Gtk

import oathguardian
from ..tools import qrcode
from ..tools import screenshot
from .datamodel import OGModel, CredentialColumns
from .utils import load_image

log = logging.getLogger(__name__)


def debug_action(obj, action, param):
    log.info('Called %s with action=%s, param=%s',
             type(obj), action, param)


class ShowAboutAction(object):
    """
    Shows the 'about' dialog.
    """
    def __init__(self, window):
        self.parent = window

    def __call__(self, action, param):
        debug_action(self, action, param)

        about = Gtk.AboutDialog(transient_for=self.parent, modal=True)
        about.set_license_type(Gtk.License.CUSTOM)
        about.set_license(oathguardian.LICENSE_TEXT)
        about.set_copyright(oathguardian.COPYRIGHT)
        about.set_comments(oathguardian.DESCRIPTION)
        about.set_authors(oathguardian.AUTHORS)
        about.set_website(oathguardian.URL)
        about.set_version(oathguardian.VERSION)
        about.set_logo_icon_name(oathguardian.NAME)
        about.present()


class ShowAddAction(object):
    """
    Shows the 'add' dialog.
    """
    @classmethod
    def name(cls):
        return 'add_entry'

    def __init__(self, appwindow):
        self.appwindow = appwindow
        self.controller = appwindow.add_controller

    def __call__(self, action, param):
        debug_action(self, action, param)

        response = self.controller.run()

        if response == 1:
            params, provider = self.controller.validate()

            self.appwindow.app.model.add_to_provider(provider, **params)
        else:
            log.debug('%s cancelled.', self.name())

        self.controller.hide()


class ScanQRAction(object):

    @classmethod
    def name(cls):
        return 'scan_qr'

    def __init__(self, application):
        self.app = application
        self.tempfile = None

    def __call__(self, action, param=None):
        debug_action(self, action, param)

        with tempfile.NamedTemporaryFile(suffix='.png', delete=False) as temp:
            with screenshot.ScreenGrabber(temp.name, self.on_success, self.on_error) as sg:
                sg.grab()

    def on_success(self, result):
        success, filename = result
        data = qrcode.decode(filename)
        log.debug('%s: decoded %s from qr code', ScanQRAction.name(), data)
        self.app.add_controller.fill(**data)

    def on_error(self, result):
        log.error(result)


class GenerateAction(object):

    @classmethod
    def name(cls):
        return 'generate'

    def __init__(self, model):
        self.model = model

    def __call__(self, view, path, column):
        self.model.generate_code(path)


class CopyTokenAction(object):
    """
    Puts the current token into the clipboard.
    """
    def __init__(self, model: OGModel, view: Gtk.TreeView):
        self.model = model
        self.view = view

    @staticmethod
    def name():
        return 'copy_password'

    def __call__(self, action, param):
        debug_action(self, action, param)

        model, tree_iter = self.view.get_selection().get_selected()
        if not tree_iter:
            log.warning('%s: no current entry: action should have been deactivated!', self.name())
            return

        current = self.model.code_from_row(model[tree_iter])
        if not current:
            log.debug('%s: no current code for row, do nothing.', self.name())

        pwd = str(current.code)
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        clipboard.set_text(pwd, -1)


class RemoveCredentialAction(object):
    """
    Removes the selected token.
    """
    def __init__(self, model: OGModel, view: Gtk.TreeView, window: Gtk.Window):
        self.window = window
        self.model = model
        self.view = view

    @staticmethod
    def name():
        return 'remove_credential'

    def __call__(self, action, param):
        debug_action(self, action, param)

        model, tree_iter = self.view.get_selection().get_selected()
        if not tree_iter:
            log.warning('%s: no current entry: action should have been deactivated!', self.name())
            return

        row = model[tree_iter]
        cred = row[CredentialColumns.CREDENTIAL]

        dialog = Gtk.MessageDialog(self.window, 0, Gtk.MessageType.WARNING,
                                   Gtk.ButtonsType.CANCEL,
                                   "Remove {}?".format(cred))
        dialog.format_secondary_text("This will permanently remove {credential} from {provider}.".format(
            credential=cred, provider=cred.provider.name
        ))
        dialog.add_button("Remove", Gtk.ResponseType.YES)
        response = dialog.run()
        dialog.destroy()

        if response != Gtk.ResponseType.YES:
            return

        try:
            self.model.remove(tree_iter)
        except Exception as e:
            log.error('Unable to remove %s', tree_iter, exc_info=e)


class RefreshProvidersAction(object):

    @classmethod
    def name(cls):
        return 'refresh-provider'

    def __init__(self, model):
        self.model = model

    def __call__(self, *args):
        self.model.refresh_providers()



# coding: utf-8

# region License
# Copyright (c) 2017,2018 Alexandre Vaissière <avaiss@fmiw.org>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# endregion

import gettext
import logging
import logging.config
import os
import signal
import sys

from gi.repository import GLib
from gi.repository import Gtk

import oathguardian
from .application import OathGuardian

log = logging.getLogger(__name__)


def main():
    logging_conf = os.path.expanduser('~/.config/oathguardian/logging.conf')
    if os.path.exists(logging_conf):
        logging.config.fileConfig(logging_conf)
    else:
        logging.basicConfig(level=logging.INFO,
                            format='%(asctime)-15s [%(threadName)-16s] %(levelname)s %(name)s: %(message)s')

    name = oathguardian.NAME

    log.info('Starting %s v%s', name, oathguardian.VERSION)

    gettext.bindtextdomain('oathguardian',
                           [os.path.join(sys.prefix, 'share', 'locale')])
    gettext.textdomain('oathguardian')

    if hasattr(GLib, "unix_signal_add"):
        GLib.unix_signal_add(GLib.PRIORITY_DEFAULT, signal.SIGINT,
                             Gtk.main_quit, None)
        GLib.unix_signal_add(GLib.PRIORITY_DEFAULT, signal.SIGTERM,
                             Gtk.main_quit, None)

    app = OathGuardian()
    app.run(sys.argv)

    log.info('Stopping %s', name)


if __name__ == '__main__':
    main()



# coding=utf-8

# region License

# Copyright (c) 2017,2018 Alexandre Vaissière <avaiss@fmiw.org>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# endregion

import logging
import time
import typing

from gi.repository import GLib
from gi.repository import Gtk

from . import utils
from .. import oath
from ..oath.providers.tools import ProviderCallback, ProviderList

log = logging.getLogger(__name__)


class CredentialColumns(utils.Columns):
    """
    Definition of the columns provided by OGModel#store store.
    """
    PROVIDER = ('Provider', object)
    CREDENTIAL = ('Credential', object)
    CODE_OBJECT = ('Code', object)
    NAME = ('Name', str)
    EXPIRATION = ('Expiration', int)
    CODE = ('Code', str)
    REMAINING_TIME = ('Remaining', int)


class ProviderColumns(utils.Columns):
    """

    """
    PROVIDER_ID = ('ID', str)
    PROVIDER = ('Provider', object)


class OGModel(ProviderCallback):
    """
    Stores the current state of oath guardian.
    """
    @classmethod
    def token_to_row(cls, token: oath.Credential, provider: oath.Provider, code=None):
        issuer = token.issuer or 'No issuer'
        assert token.name is not None

        expiration = code and code.expiry or 0
        return (provider, token, None,
                "<big>{}</big>\n{}".format(GLib.markup_escape_text(issuer),
                                           GLib.markup_escape_text(token.name)),
                expiration,
                cls._format_code(code=code),
                0)

    @classmethod
    def token_from_row(cls, row) -> oath.Credential:
        return row[CredentialColumns.CREDENTIAL]

    @classmethod
    def code_from_row(cls, row) -> oath.Code:
        return row[CredentialColumns.CODE_OBJECT]

    @classmethod
    def _compute_expiration(cls, token: oath.Credential, row, timestamp=None):
        code = cls.code_from_row(row)
        if not code:
            return 0

        expiration = row[CredentialColumns.EXPIRATION]
        if not expiration:
            return 0

        if expiration < 0:
            return 100

        if not timestamp:
            timestamp = int(time.time())

        if expiration > timestamp:
            remaining_time = expiration - timestamp
            log.debug('For token %s expiring at %d: remaining %ds',
                      token.name, expiration, remaining_time)
            return min(100, 100 * remaining_time / code.period)
        else:
            return 0

    @classmethod
    def _format_code(cls, code: oath.Code=None):
        if code:
            formatter = oath.new_human_formatter(code.digits)
            code = '<b>{}</b>'.format(GLib.markup_escape_text(formatter(code.code)))
        else:
            code = '______'

        return '<big><tt>{}</tt></big>'.format(code)

    def __init__(self, settings):
        self._credentials = Gtk.ListStore(*CredentialColumns.column_types())
        self._providers_store = Gtk.ListStore(*ProviderColumns.column_types())

        self._providers_list = ProviderList(settings)
        self._providers_list.add_listener(self)
        self._providers_list.refresh()

        self.started = True
        # Not sure it is worth it to do this every demi-second...
        # self._timer_source = GLib.timeout_add(interval=500, function=self.refresh_providers)
        self._check_expiration_source = GLib.timeout_add_seconds(interval=1, function=self._check_expiration)

    @property
    def providers(self) -> Gtk.ListStore:
        return self._providers_store

    @property
    def credentials(self) -> Gtk.ListStore:
        return self._credentials

    @property
    def default_provider_id(self) -> typing.Optional[str]:
        # FIXME default provider should be taken from user configuration
        if len(self._providers_store):
            return self._providers_store[0][0]
        return None

    def _unlock(self, provider: oath.Provider):
        if not provider.is_locked():
            return True

        return provider.unlock()

    def add_from_provider(self, provider: oath.Provider):
        self._providers_store.append([provider.id, provider])

        if not self._unlock(provider):
            # FIXME Could not unlock the provider: abort.
            return

        for token in provider.list():
            row = self.token_to_row(token, provider)
            log.debug("Adding %s -> %s from %s", token, row, provider)
            self._credentials.append(row=row)

    def add_to_provider(self, provider: oath.Provider, **kwargs):
        assert provider

        cred = provider.add(**kwargs)
        self._credentials.append(row=self.token_to_row(cred, provider))

    def remove_from_provider(self, provider: oath.Provider):
        for row in self._credentials:
            if row[CredentialColumns.PROVIDER] == provider:
                self._credentials.remove(row.iter)
        for row in self._providers_store:
            if row[ProviderColumns.PROVIDER] == provider:
                self._providers_store.remove(row.iter)

    def _check_expiration(self):
        """
        This method is periodically ran from glib thread in order to update the remaining time on each token.
        :return bool: True, to keep running.
        """
        timestamp = int(time.time())

        log.debug('Check expiration at time %d', timestamp)

        for row in self._credentials:
            token = self.token_from_row(row)
            remaining_percent = self._compute_expiration(token, row, timestamp=timestamp)
            row[CredentialColumns.REMAINING_TIME] = remaining_percent
            if not remaining_percent:
                row[CredentialColumns.CODE_OBJECT] = None
                row[CredentialColumns.CODE] = OGModel._format_code(code=None)

        return True

    def generate_code(self, path):
        row = self._credentials[path]
        if row:
            token = self.token_from_row(row)
            code = token.compute()
            log.debug('Generated %s for token %s', code, token)
            self._credentials[path][CredentialColumns.CODE_OBJECT] = code
            self._credentials[path][CredentialColumns.CODE] = self._format_code(code=code)
            self._credentials[path][CredentialColumns.EXPIRATION] = code.expiry

    def remove(self, tree_iter):
        row = self._credentials[tree_iter]
        if row:
            cred = self.token_from_row(row)
            log.info('Removing token %s', cred)

            cred.provider.remove(cred)

            self._credentials.remove(tree_iter)

    def refresh_providers(self):
        self._providers_list.refresh()
        return True

    def on_added_provider(self, provider: oath.Provider):
        log.debug('New provider added: %s (%s)', provider.name, provider.id)
        self.add_from_provider(provider)

    def on_removed_provider(self, provider: oath.Provider):
        log.debug('Provider removed: %s (%s)', provider.name, provider.id)
        self.remove_from_provider(provider)



# coding: utf-8

# region License

# Copyright (c) 2018 Alexandre Vaissière <avaiss@fmiw.org>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# endregion

"""

"""

import base64
import typing

key_types = typing.Union[str, bytes, bytearray]
"""Accepted types for key in generate_totp, generate_hotp methods. str means 'b32 encoded'."""


def b32_normalize(key: str) -> str:
    """
    Web sites often gives the secret with spaces between 4-char words. This method normalizes
    the received key into one parseable by base64, and handles padding.

    :param key: the original key.
    :return: a normalized b32 string.
    """
    # This piece of really clever code was taken from ykman.util.parse_b32_key
    key = key.upper().replace(' ', '')
    # if key length is not a multiple of 8, pad with '=' chars.
    key += '=' * (-len(key) % 8)
    return key


def to_binary_key(key: key_types) -> bytes:
    """
    Converts a key to its bytes representation.

    :param key: a key, either as a base32 encoded string, or a byte array.
    :return: the binary key.
    """
    if isinstance(key, base64.bytes_types):
        return key
    else:
        return base64.b32decode(b32_normalize(key))


def to_b32_key(key: key_types) -> str:
    """
    Converts a key to its string representation, in base32.

    :param key: a key, either as a base32 encoded string, or a byte array.
    :return: the base32 encoded key.
    """
    if isinstance(key, base64.bytes_types):
        return base64.b32encode(key).decode('ascii')
    return b32_normalize(key)



# coding: utf-8

# region License

# Copyright (c) 2018 Alexandre Vaissière <avaiss@fmiw.org>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# endregion

"""
Utilities to format TOTP/HOTP codes in an human readable way.
"""

import typing


formatter_type = typing.Callable[[int], str]
"""Type of a formatter method."""


def new_default_formatter(digits: int=6) -> formatter_type:
    """
    Returns a method that computes the stringified version of an HOTP/TOTP code using digits, left-padded with 0 to
    reach the requested number of digits.
    :param digits: number of requested digits.
    :return: formatter method.
    """
    def format_code(hotp: int) -> str:
        return '{:0{digits}d}'.format(hotp, digits=digits)
    return format_code


def new_steam_formatter() -> formatter_type:
    """
    Returns a method that computes the stringified version of an HOTP/TOTP in steam expected format.
    :return: formatter method.
    """
    steam_alphabet = '23456789BCDFGHJKMNPQRTVWXY'
    base = len(steam_alphabet)

    def format_code(hotp: int) -> str:
        chars = []
        for i in range(5):
            hotp, mod = divmod(hotp, base)
            chars.append(steam_alphabet[mod])
        return ''.join(chars)

    return format_code


def new_human_formatter(digits: int=6) -> formatter_type:
    """
    Formats the code by blocks of 3 or 4 characters, making them simpler
    to retain.

    :param digits:
    :return:

    >>> f = new_human_formatter(digits=8)
    >>> f(12345678)
    '1234 5678'
    >>> f(1234567)
    '0123 4567'
    >>> f = new_human_formatter()
    >>> f(12345678)
    '1234 5678'
    >>> f(1234567)
    '123 4567'
    >>> f(123456)
    '123 456'
    >>> f(12345)
    '012 345'
    """
    def format_code(hotp: int) -> str:
        s = '{:0{digits}d}'.format(hotp, digits=digits)

        length = len(s)
        if length < 4:
            return s

        sep = 4
        if length % 4 != 0 and length % 3 == 0:
            sep = 3

        splitted = [s[max(0, i - sep):i] for i in range(length, 0, -sep)]
        return ' '.join(reversed(splitted))

    return format_code



# coding=utf-8

# region License

# Copyright (c) 2017,2018 Alexandre Vaissière <avaiss@fmiw.org>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# endregion

from gi.repository import GLib
from gi.repository import Gio


class ScreenGrabber(object):
    def __init__(self, filename, result_handler, error_handler):
        self.filename = filename
        self.result_handler = result_handler
        self.error_handler = error_handler

    def __enter__(self):
        self._bus = Gio.bus_get_sync(Gio.BusType.SESSION, None)
        self._proxy = Gio.DBusProxy.new_sync(self._bus, Gio.DBusProxyFlags.NONE, None,
                                             'org.gnome.Shell.Screenshot',
                                             '/org/gnome/Shell/Screenshot',
                                             'org.gnome.Shell.Screenshot',
                                             None)
        self._main_loop = GLib.MainLoop()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def _on_error(self, obj, result, user_data):
        self.error_handler(result)

    def _on_success(self, obj, result, user_data):
        self.result_handler(result)

    def _screenshot(self, proxy, result, user_data):
        x, y, w, h = result
        proxy.ScreenshotArea('(iiiibs)',
                             x, y, w, h, False, user_data,
                             user_data=user_data,
                             result_handler=self._on_success, error_handler=self._on_error)

    def grab(self):
        self._proxy.SelectArea(timeout=5000,
                               result_handler=self._screenshot,
                               error_handler=self._on_error,
                               user_data=self.filename)



# coding=utf-8

# region License

# Copyright (c) 2017,2018 Alexandre Vaissière <avaiss@fmiw.org>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# endregion

"""
Yubikey 4 provider implementation.
"""

import logging
import time
import typing

import smartcard.Exceptions

import ykman.oath as ykoath
from ykman import driver_ccid

from . import Provider, ProviderCapability, Credential, OATHType, HashAlgo
from .. import errors
from .. import config
from ... import oath

from oathguardian.tools import gnupg

log = logging.getLogger(__name__)


class YubikeyOath(Provider):
    """
    Oath provider using YKOATH protocol (yubikey 4).
    """

    def __init__(self, settings: config.Config, driver: driver_ccid.CCIDDriver):
        self.settings = settings
        self.driver = driver
        self.controller = ykoath.OathController(driver)

    @property
    def name(self):
        return '{} {}'.format(self.driver.key_type.value, self.driver.serial)

    @property
    def id(self) -> str:
        return 'yubikey:{}'.format(self.driver.serial)

    @property
    def capabilities(self) -> ProviderCapability:
        return ProviderCapability.TOTP | ProviderCapability.HOTP | ProviderCapability.LOCK

    @classmethod
    def _to_yubikey_credential(cls, orig: oath.Credential) -> ykoath.Credential:
        if isinstance(orig, YubiCredential):
            return orig._yubi_cred
        raise ValueError('credential {} does not come from yubikey'.format(orig))

    def remove(self, cred: Credential):
        self.controller.delete(YubikeyOath._to_yubikey_credential(cred))

    def add(self, oath_type: OATHType, label: str, secret: oath.key_types,
            issuer: str = None, algorithm: HashAlgo = HashAlgo.SHA1, digits: int = 6,
            counter: int = None, period: int = 30, **kwargs):
        if oath_type not in (oath.OATHType.TOTP, oath.OATHType.HOTP):
            raise errors.UnsupportedOATHTypeError('{} does not support {} types'.format(self, oath_type))

        key_bytes = oath.to_binary_key(secret)

        if not counter:
            counter = 0

        data = ykoath.CredentialData(key_bytes, issuer, label, oath_type=ykoath.OATH_TYPE[oath_type.name],
                                     algorithm=ykoath.ALGO[algorithm.name], digits=digits, period=period,
                                     counter=counter, touch=kwargs.get('require_touch', True))

        log.info('Adding %s to %s', data.make_key(), self)

        try:
            cred = self.controller.put(data)
            return self._wrap(cred)
        except ykoath.APDUError as e:
            if e.sw == ykoath.SW.NO_SPACE:
                raise errors.NoSpaceLeftError('No space left on device.')
            elif e.sw == ykoath.SW.COMMAND_ABORTED:
                raise errors.NoSpaceLeftError('The command failed. Do you have enough space on the device?')
            else:
                raise

    def list(self):
        return [self._wrap(cred) for cred in self.controller.list()]

    def _wrap(self, cred: ykoath.Credential) -> Credential:
        return YubiCredential(self, cred)

    def is_locked(self):
        return self.controller.locked

    def unlock(self):
        with gnupg.pinentry(self.settings.pinentry_executable()) as pinentry:
            pinentry.set_option('allow-external-password-cache')
            pinentry.set_keyinfo('oathguardian: ' + self.id)

            for attempt in range(3):
                try:
                    password = pinentry.ask_pin('Please enter passphrase for {}'.format(self.id))
                except gnupg.AssuanError as e:
                    log.error("Could not get pin from user: 0x%x %s", e.code, e)
                    return

                key = self.controller.derive_key(password)
                try:
                    self.controller.validate(key)
                    return True
                except Exception as e:
                    log.error('Unable to unlock: %s', e, exc_info=e)
                    pinentry.set_error('Invalid password. Remaining {} attempts.'.format(2 - attempt))

    def __del__(self):
        log.info('Releasing %s', self.name)
        del self.controller._driver

    def __str__(self):
        return self.id

    def __repr__(self):
        return 'YubikeyOath<{}>'.format(self.id)


class YubiCredential(oath.Credential):
    def __init__(self, provider: YubikeyOath, yubi_cred: ykoath.Credential):
        super(YubiCredential, self).__init__(yubi_cred.name, issuer=yubi_cred.issuer)
        self._yubi_cred = yubi_cred
        self._provider = provider
        self._type = oath.OATHType[yubi_cred.oath_type.name]

    @property
    def provider(self):
        return self._provider

    @property
    def type(self) -> OATHType:
        return self._type

    def _compute0(self, **kwargs) -> typing.Tuple[int, int, int, int]:
        code = self._provider.controller.calculate(self._yubi_cred)

        expiry = code.valid_to
        if self._yubi_cred.oath_type == ykoath.OATH_TYPE.HOTP:
            expiry = -1

        # FIXME does not work every time if the credential is named steam.
        return int(code.value), len(code.value), expiry, self._yubi_cred.period


class _ProviderDescriptor(object):
    def __init__(self, settings, drv: driver_ccid.CCIDDriver):
        self._settings = settings
        self._driver = drv

    @property
    def id(self):
        return 'yubikey:{}'.format(self._driver.serial)

    def open(self):
        return YubikeyOath(self._settings, self._driver)

    def __del__(self):
        del self._driver


class _YubikeyFinder(object):
    """Finds currently plugged in yubikeys and opens them. This instance stores the currently
    known keys, so that they are not opened several times; and permits consecutive calls to
    :func:`update_readers` to return a list of added and removed keys.
    """
    def __init__(self, settings):
        """

        :param config.Config settings: oathguardian configuration.
        """
        self.settings = settings
        self._cards = {}
        self._attempts = 3

    def _open_reader(self, reader):
        attempt = 0
        while True:
            sleep_time = (attempt + 1) * 0.1
            try:
                conn = reader.createConnection()
                conn.connect()
                return driver_ccid.CCIDDriver(conn, reader.name)
            except smartcard.Exceptions.CardConnectionException as e:
                log.error("Could not open CCID: %s", str(e), exc_info=e)
                if attempt < self._attempts:
                    attempt += 1
                    gnupg.kill_scdaemon(socket_path=self.settings.gpg_agent_path())
                    time.sleep(sleep_time)
                    continue
                raise

    def update_readers(self):
        log.debug('Updating readers, known being %s', self._cards.keys())

        current_readers = {reader.name: reader for reader in driver_ccid._list_readers()
                           if reader.name.lower().startswith('yubico yubikey')}

        known_names = set(self._cards.keys())
        current_names = set(current_readers.keys())

        # handle removed readers
        removed = []
        removed_names = known_names - current_names
        for name in removed_names:
            desc = self._cards.pop(name)
            removed.append(desc.id)
            del desc

        added = []
        added_names = current_names - known_names
        for name in added_names:
            try:
                driver = self._open_reader(current_readers[name])
                desc = _ProviderDescriptor(self.settings, driver)
                self._cards[name] = desc
                added.append(desc)
            except smartcard.Exceptions.CardConnectionException as e:
                log.warning('Could not open %s: %s', name, str(e), exc_info=e)
                continue

        log.debug('New cards added: %s, removed: %s', added, removed)

        return added, removed


def get_finder(settings):
    """
    :param config.Config settings:
    :return:
    """
    # Try to disable this code in case pcscd is not enabled or launched or does not work.

    from smartcard.pcsc.PCSCContext import PCSCContext
    from smartcard.pcsc.PCSCExceptions import EstablishContextException
    try:
        PCSCContext().getContext()
    except EstablishContextException as e:
        log.error('Error while initializing pcsc: %s. Maybe pcscd is not installed on your system! ' +
                  'Yubikey storage disabled', str(e), exc_info=e)
    except Exception as e:
        log.error('Error while initializing pcsc: %s. Yubikey storage disabled',
                  str(e), exc_info=e)
    else:
        return _YubikeyFinder(settings)



# coding=utf-8

# region License

# Copyright (c) 2017,2018 Alexandre Vaissière <avaiss@fmiw.org>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# endregion

import abc
import enum
import functools
import typing

from collections import namedtuple

from ... import oath


@enum.unique
class OATHType(enum.Enum):
    """Algorithm types."""
    TOTP = 'totp'
    HOTP = 'hotp'


@enum.unique
class ProviderCapability(enum.IntFlag):
    TOTP = enum.auto()
    HOTP = enum.auto()
    LOCK = enum.auto()


@enum.unique
class HashAlgo(enum.Enum):
    """Supported hash algorithms."""
    SHA1 = 'sha1'
    SHA256 = 'sha256'
    SHA512 = 'sha512'


Code = namedtuple('Code', ['code', 'representation', 'digits', 'expiry', 'period'])
Code.__doc__ = """Result of a TOTP/HOTP computation.
:param int code: actual result.
:param int digits: number of digits required for code (for a formatter to properly pad with zeroes if needed).
:param int expiry: timestamp at which this code expires.
:param int period: validity duration of this code.
"""


@functools.total_ordering
class Credential(metaclass=abc.ABCMeta):
    """
    Object identifying a credential stored in a provider.
    """
    def __init__(self, name: str, issuer: str=None):
        self._name = name
        self._issuer = issuer

    @property
    @abc.abstractmethod
    def provider(self) -> 'Provider':
        pass

    @property
    def issuer(self) -> str:
        return self._issuer

    @property
    def name(self) -> str:
        return self._name

    @property
    @abc.abstractmethod
    def type(self) -> OATHType:
        pass

    def compute(self, **kwargs) -> Code:
        code, digits, expiry, period = self._compute0(**kwargs)
        formatter = kwargs.pop('formatter', oath.new_default_formatter(digits))
        return Code(code, formatter(code), digits, expiry, period)

    @abc.abstractmethod
    def _compute0(self, **kwargs) -> typing.Tuple[int, int, int, int]:
        raise NotImplementedError

    def __lt__(self, other):
        a = (self.issuer or '', self.name)
        b = (other.issuer or '', other.name)
        return a < b

    def __str__(self):
        return '{}:{}'.format(self.issuer or '', self.name)


class Provider(metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def id(self) -> str:
        """Unique identifier for the provider. Preferred way to build this id is by using a pattern
        {provider_type}:{provider_key}.

        :return str: provider identifier.
        """
        pass

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Display name
        :return str: display name
        """
        pass

    @property
    @abc.abstractmethod
    def capabilities(self) -> ProviderCapability:
        pass

    @abc.abstractmethod
    def list(self) -> typing.List[Credential]:
        pass

    @abc.abstractmethod
    def add(self, oath_type: OATHType, label: str, secret: oath.key_types,
            issuer: str=None, algorithm: HashAlgo=HashAlgo.SHA1, digits: int=6,
            counter: int=None, period: int=30, **kwargs) -> Credential:
        pass

    @abc.abstractmethod
    def remove(self, cred: Credential):
        """Removes the given credential from this storage.
        :param Credential cred: credential to remove.
        :throw ValueError: if the give credential i
        """
        pass

    @abc.abstractmethod
    def is_locked(self) -> bool:
        """Tells if the current storage is locked. If locked, a call to unlock method is required
        before being able to do any other operation of the storage.
        :return bool: True if this storage is locked.
        """
        pass

    @abc.abstractmethod
    def unlock(self) -> bool:
        pass



# coding: utf-8

# Copyright (c) 2018 Alexandre Vaissière <avaiss@fmiw.org>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.


"""

"""

import configparser
import logging
import os

log = logging.getLogger(__name__)


class Config(object):
    def __init__(self):
        self._config = configparser.ConfigParser()

    def load_config(self):
        filename = os.path.expanduser('~/.config/oathguardian/oathguardian.conf')
        if os.path.exists(filename):
            log.info('Loading configuration from %s', filename)
            self._config.read(filenames=[filename], encoding='utf-8')

    def gpg_agent_path(self):
        default_path = '/run/user/{uid}/gnupg/S.gpg-agent'.format(uid=os.getuid())
        if 'gnupg' in self._config:
            return self._config['gnupg'].get('gpg_agent_path', default_path)
        return default_path

    def pinentry_executable(self):
        default_executable = 'pinentry'
        if 'gnupg' in self._config:
            return self._config['gnupg'].get('pinentry', default_executable)
        return default_executable



# coding=utf-8

# region License

# Copyright (c) 2017,2018 Alexandre Vaissière <avaiss@fmiw.org>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# endregion

import logging
import time
import typing

import secretstorage
import secretstorage.collection

import oathguardian
from oathguardian import oath
from oathguardian.oath import errors
from oathguardian.oath.providers import ProviderCapability

log = logging.getLogger(__name__)


class LocalProvider(oath.Provider):
    """
    The local provider stores secret in gnome keyring (or any application
    implementing the freedesktop secret interface).
    """
    def __init__(self, collection: secretstorage.Collection):
        self.collection = collection
        self._id = 'local:' + collection.get_label()

    @staticmethod
    def make_name(label: str, issuer: str=None) -> str:
        if not issuer:
            return label

        if label.startswith(issuer + ':'):
            return label

        return issuer + ':' + label

    def add(self, oath_type: oath.OATHType, label: str, secret: oath.key_types,
            issuer: str=None, algorithm: oath.HashAlgo=oath.HashAlgo.SHA1, digits: int=6,
            counter: int=None, period: int=30, **kwargs) -> oath.Credential:
        if oath_type not in (oath.OATHType.TOTP, oath.OATHType.HOTP):
            raise errors.UnsupportedOATHTypeError('{} does not support {} types'.format(self, oath_type))

        key_bytes = oath.to_b32_key(secret).encode('utf-8')

        options = {
            'oath_type': oath_type.name.lower(),
            'algorithm': algorithm.value,
            'issuer': issuer,
            'digits': str(digits),
            'created-by': oathguardian.NAME,
        }

        if oath_type == oath.OATHType.TOTP:
            options.update({'period': str(period)})
        elif oath_type == oath.OATHType.HOTP:
            options.update({'counter': str(counter)})

        name = self.make_name(label, issuer)

        log.info('Adding %s -> %s to %s collection.', name, options, self.collection.get_label())

        item = self.collection.create_item(name, options, key_bytes)
        return LocalCredential(self, item)

    def list(self):
        if self.collection.is_locked():
            raise Exception('Collection is locked!')

        def is_cred(item):
            attrs = item.get_attributes()
            return oathguardian.NAME == attrs.get('created-by') and attrs.get('oath_type') in ('totp', 'hotp')

        return [self._wrap(item) for item in self.collection.get_all_items() if is_cred(item)]

    def _wrap(self, item: secretstorage.Item) -> oath.Credential:
        return LocalCredential(self, item)

    def unlock(self):
        self.collection.unlock(self._on_unlock)
        return not self.collection.is_locked()

    def _on_unlock(self, dismissed, items):
        if not dismissed:
            log.info('Unlocked collection %s (containing %d items)', self.id, len(items))
        else:
            log.info('User dismissed the password prompt.')

    def remove(self, cred):
        if cred.provider != self:
            raise ValueError('credential {} does not come from {}'.format(cred, self.id))
        cred._item.delete()

    @property
    def capabilities(self):
        return ProviderCapability.TOTP | ProviderCapability.LOCK

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str:
        return '{collection_name} keyring'.format(collection_name=self.collection.get_label())

    def is_locked(self):
        return self.collection.is_locked()


class LocalCredential(oath.Credential):

    def __init__(self, provider, item: secretstorage.Item):
        attrs = item.get_attributes()
        label, issuer = oath.uri.normalize_name(item.get_label(), attrs.get('issuer'))
        super(LocalCredential, self).__init__(label, issuer=issuer)

        self._digits = int(attrs.get('digits'))
        self._provider = provider
        self._item = item
        self._algorithm = oath.HashAlgo[attrs.get('algorithm').upper()]
        self._type = oath.OATHType[attrs.get('oath_type').upper()]
        self._period = int(attrs.get('period', 0))
        self._counter = int(attrs.get('counter', 0))

    @property
    def provider(self):
        return self._provider

    @property
    def type(self) -> oath.OATHType:
        return self._type

    def _fetch_key(self) -> bytes:
        return self._item.get_secret()

    def _compute0(self, **kwargs) -> typing.Tuple[int, int, int, int]:
        key = self._fetch_key()

        if self.type == oath.OATHType.TOTP:
            timestamp = int(time.time())
            totp = oath.generate_totp(key, timestamp=timestamp, period=self._period, digits=self._digits,
                                      hash_algo=self._algorithm.value)
            return totp, self._digits, (timestamp//self._period + 1) * self._period, self._period

        if self.type == oath.OATHType.HOTP:
            hotp = oath.generate_hotp(key, counter=self._counter, digits=self._digits, hash_algo=self._algorithm.value)

            # Update counter in our item.
            attrs = self._item.get_attributes()
            self._counter += 1
            attrs['counter'] = str(self._counter)
            self._item.set_attributes(attrs)

            return hotp, self._digits, -1, -1


def find_providers(known=None):
    if 'oathguardian' in known:
        return

    bus = secretstorage.dbus_init()
    for collection in secretstorage.get_all_collections(bus):
        if 'oathguardian' == collection.get_label():
            yield LocalProvider(collection)
            return

    collection = secretstorage.create_collection(bus, 'oathguardian')
    yield LocalProvider(collection)


def default() -> LocalProvider:
    """
    Gets a provider that store all credentials on secret-service default collection.
    :return: the default provider.
    """
    bus = secretstorage.dbus_init()
    collection = secretstorage.get_default_collection(bus)
    return LocalProvider(collection)



# coding: utf-8

# region License

# Copyright (c) 2017,2018 Alexandre Vaissière <avaiss@fmiw.org>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# endregion


import logging

from gi.repository import GLib
from gi.repository import Gio
from gi.repository import Gtk

import oathguardian
from . import controllers
from . import actions
from . import datamodel
from .utils import load_uifile
from oathguardian.oath import config

log = logging.getLogger(__name__)


class MainWindow(object):
    """Wrapper around the main application window."""
    def __init__(self, application, title=None):
        super(MainWindow, self).__init__()
        self.app = application

        widget_tree = load_uifile('oathguardian.xml')
        self.window = widget_tree.get_object('mainwindow')
        self.__configure_window(application, title)

        self.tokens_view = widget_tree.get_object('tokens_view')
        self.__init_treeview(self.tokens_view, self.app.model)

        self.add_controller = controllers.AddDialog(self, widget_tree)

        action = Gio.SimpleAction.new("about", None)
        action.connect("activate", actions.ShowAboutAction(self.window))
        self.app.add_action(action)

        action = Gio.SimpleAction.new('refresh', None)
        action.connect('activate', actions.RefreshProvidersAction(self.app.model))
        self.app.add_action(action)
        self.app.add_accelerator('<Control>R', 'app.refresh', None)

        action = Gio.SimpleAction.new("add", None)
        action.connect("activate", actions.ShowAddAction(self))
        self.app.add_action(action)
        self.app.add_accelerator('<Control>N', 'app.add', None)

        action = Gio.SimpleAction.new("copy", None)
        action.connect("activate", actions.CopyTokenAction(self.app.model, self.tokens_view))
        self.app.add_action(action)
        self.app.add_accelerator('<Control>C', 'app.copy', None)

        action = Gio.SimpleAction.new("remove", None)
        action.connect("activate", actions.RemoveCredentialAction(self.app.model, self.tokens_view, self.window))
        self.app.add_action(action)
        self.app.add_accelerator('Delete', 'app.remove', None)

        # This will be in the windows group and have the "win" prefix
        max_action = Gio.SimpleAction.new_stateful("maximize", None, GLib.Variant.new_boolean(False))
        max_action.connect("change-state", self.on_maximize_toggle)
        self.window.add_action(max_action)

        # Keep it in sync with the actual state
        self.window.connect("notify::is-maximized",
                            lambda obj, pspec: max_action.set_state(GLib.Variant.new_boolean(obj.props.is_maximized)))

    def __configure_window(self, application, title):
        self.window.set_application(application)
        if title:
            self.window.set_title(title)

    def __init_treeview(self, view: Gtk.TreeView, model: datamodel.OGModel):
        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn(datamodel.CredentialColumns.NAME.display_name,
                                    cell_renderer=renderer, markup=datamodel.CredentialColumns.NAME)
        view.append_column(column)

        renderer = Gtk.CellRendererProgress()
        renderer.set_orientation(Gtk.Orientation.VERTICAL)
        renderer.set_property('inverted', True)
        renderer.set_property('text', '')
        column = Gtk.TreeViewColumn(datamodel.CredentialColumns.REMAINING_TIME.display_name,
                                    cell_renderer=renderer, value=datamodel.CredentialColumns.REMAINING_TIME)
        view.append_column(column)

        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn(datamodel.CredentialColumns.CODE.display_name,
                                    cell_renderer=renderer, markup=datamodel.CredentialColumns.CODE)
        view.append_column(column)

        view.connect('row-activated', actions.GenerateAction(model))
        view.set_model(model.credentials)
        view.set_search_column(datamodel.CredentialColumns.NAME)
        view.set_enable_search(True)
        view.set_search_equal_func(self._search)

    @staticmethod
    def _search(model, column, search, row_iter, user_data=None):
        value = model[row_iter][column]
        return value and search and search not in value

    def present(self):
        self.window.present()

    def on_maximize_toggle(self, action, value):
        action.set_state(value)
        if value.get_boolean():
            self.window.maximize()
        else:
            self.window.unmaximize()


class OathGuardian(Gtk.Application):
    """Application."""
    def __init__(self, *args, **kwargs):
        super(Gtk.Application, self).__init__(*args, application_id="org.fmiw.oathguardian",
                                              #flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE,
                                              **kwargs)
        GLib.set_application_name(oathguardian.NAME)
        GLib.set_prgname(oathguardian.NAME)

        self.window = None
        self.provider = None
        self.model = None
        self.config = config.Config()

    def do_startup(self):
        Gtk.Application.do_startup(self)

        self.config.load_config()

        action = Gio.SimpleAction.new("about", None)
        action.connect("activate", self._on_about)
        self.add_action(action)

        action = Gio.SimpleAction.new("quit", None)
        action.connect("activate", self._on_quit)
        self.add_action(action)

        builder = load_uifile('oathguardian_menu.xml')
        self.set_app_menu(builder.get_object("app-menu"))

        self.model = datamodel.OGModel(self.config)

    def do_activate(self):
        # We only allow a single window and raise any existing ones
        if not self.window:
            # Windows are associated with the application
            # when the last one is closed the application shuts down
            self.window = MainWindow(application=self, title=oathguardian.TITLE)

        self.window.present()

    def _on_about(self, action, param):
        about_dialog = Gtk.AboutDialog(transient_for=self.window, modal=True)
        about_dialog.present()

    def _on_quit(self, action, param):
        self.quit()



# coding=utf-8

# region License

# Copyright (c) 2017,2018 Alexandre Vaissière <avaiss@fmiw.org>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# endregion

NAME = 'oathguardian'
TITLE = 'OathGuardian'
VERSION = '0.1.0'
DESCRIPTION = 'TOTP generator for Gnome'
AUTHORS = [
    'Alexandre Vaissière <avaiss@fmiw.org>',
]
LICENSE_NAME = 'OpenBSD'
URL = 'https://framagit.org/avaiss/oathguardian'
COPYRIGHT = 'Copyright (c) 2017,2018 Alexandre Vaissière <avaiss@fmiw.org>'
LICENSE_TEXT = '''
Permission to use, copy, modify, and distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
'''



# coding: utf-8

# region License

# Copyright (c) 2017,2018 Alexandre Vaissière <avaiss@fmiw.org>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# endregion

import logging

try:
    from pyzbar import pyzbar
    from PIL import Image
    decoding_supported = True
except:
    decoding_supported = False

from oathguardian import oath

log = logging.getLogger(__name__)

_DEFAULTS = {
    'algorithm': 'SHA1',
    'digits': 6,
    'period': 30,
}


def decode(filename):
    with Image.open(filename) as image:
        data = pyzbar.decode(image)

    if not data:
        return {}

    decoded = data[0]
    if 'QRCODE' != decoded.type:
        return {}

    url = decoded.data.decode('utf-8')

    log.debug('Decoded %s', url)

    return oath.parse_uri(url)



# coding: utf-8

# region License

# Copyright (c) 2018 Alexandre Vaissière <avaiss@fmiw.org>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# endregion

"""

"""

from . import Provider
from . import local, yubikey


class ProviderCallback(object):
    """

    """
    def on_added_provider(self, provider: Provider):
        pass

    def on_removed_provider(self, provider: Provider):
        pass


class ProviderList(object):
    """Contains the list of current providers.
    Each call to refresh method query the system for current list of plugged yubikeys.
    """
    def __init__(self, settings):
        self._contents = {}
        self._callbacks = []

        # By default we load the secret storage default one.
        dflt = local.default()
        self._contents[dflt.id] = dflt

        self._yubikey_finder = yubikey.get_finder(settings)

    def __iter__(self):
        return iter(self._contents.values())

    def add_listener(self, listener: ProviderCallback) -> bool:
        if listener in self._callbacks:
            return False
        self._callbacks.append(listener)

        # send current contents
        for item in self._contents.values():
            listener.on_added_provider(item)

        return True

    def remove_listener(self, listener: ProviderCallback):
        self._callbacks.remove(listener)

    def _emit_added(self, provider):
        for callback in self._callbacks:
            callback.on_added_provider(provider)

    def _emit_removed(self, provider):
        for callback in self._callbacks:
            callback.on_removed_provider(provider)

    def refresh(self):
        if not self._yubikey_finder:
            return

        added, removed = self._yubikey_finder.update_readers()

        # notify of removed providers
        for id_removed in removed:
            prov = self._contents.pop(id_removed)
            self._emit_removed(prov)

        # creates new providers
        for desc in added:
            prov = desc.open()
            self._contents[desc.id] = prov
            self._emit_added(prov)



# coding: utf-8

# region License
# Copyright (c) 2017,2018 Alexandre Vaissière <avaiss@fmiw.org>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# endregion

"""
Tools for talking talk to gpg tools (gpg-agent and pinentry).
"""

import contextlib
import logging
import io
import os
import re
import socket
import subprocess
import typing

log = logging.getLogger(__name__)


_MAPPING = (('\n', '%0A'), ('\r', '%0D'), ('%', '%25'))


class AssuanError(Exception):
    """
    Error raised when a command sent to gpg-agent or pinentry returned with an error.
    """
    ERROR_PATTERN = re.compile('(?P<errorcode>\d+) (?P<message>.*)')

    def __init__(self, line):
        self.code, self.message = self._parse(line)

    @classmethod
    def _parse(cls, line: str):
        m = cls.ERROR_PATTERN.match(line)
        if m:
            return int(m.group('errorcode')), m.group('message')
        return 0, line


class AssuanProtocolError(Exception):
    """

    """
    pass


class Assuan(object):
    """
    A base class for sending commands and receiving responses to a process supporting
    an assuan protocol.
    """
    MAX_LINE_LENGTH = 1000
    RESPONSE_PATTERN = re.compile('^(?P<type>D|S|#|OK|ERR)(?:\s+(?P<data>.*))?$')

    def __init__(self, name, reader, writer):
        self.name = name
        self.input = reader
        self.output = writer
        self._log = logging.getLogger('{module}.{cls}.{cls}-{name}'.format(module=__name__,
                                                                           cls=Assuan.__name__,
                                                                           name=name))
        self._readline()

    def _log_input(self, line):
        self._log.info('<<< %s', line)

    def _log_output(self, line):
        self._log.info('>>> %s', line)

    def _readline(self, log_data: bool=True) -> typing.Tuple[str, str]:
        line = self.input.readline()
        if not line:
            raise AssuanProtocolError('EOF')

        line = line.rstrip()

        if len(line) > self.MAX_LINE_LENGTH:
            self._log_input(line)
            raise AssuanProtocolError('line too long')

        matcher = self.RESPONSE_PATTERN.match(line)
        if not matcher:
            raise AssuanProtocolError('Invalid response {}'.format(line))

        msg_type = matcher.group('type')
        if msg_type == 'D' and not log_data:
            self._log_input('D <hidden_data>')
        else:
            self._log_input(line)

        return msg_type, matcher.group('data') or ''

    def _responses(self, log_data: bool=True) -> typing.Generator[typing.Tuple[str, str], None, None]:
        while True:
            resp_type, resp_data = self._readline(log_data=log_data)
            if resp_type == '#':
                continue
            yield resp_type, self._decode(resp_data)
            if resp_type in ('OK', 'ERR'):
                return

    @staticmethod
    def _encode(s: str):
        tr = str.maketrans({decoded: encoded for decoded, encoded in _MAPPING})
        return s.translate(tr)

    @staticmethod
    def _decode(s: str):
        # Cannot use translate here as encoded is more than 1 char
        mapping = {encoded: decoded for decoded, encoded in _MAPPING}
        pattern = re.compile('|'.join(encoded for _, encoded in _MAPPING))
        return pattern.sub(lambda m: mapping.get(m.group(0), ''), s)

    def set_option(self, name, value=None):
        """
        Set the given option, with potential value.

        This is a shortcut to sending command SETOPTION name value.

        :param name: name of the option to set.
        :param value: optional value for the option
        :return:
        """
        args = [name]
        if value:
            args.append(str(value))
        return self.send('OPTION', args)

    def send(self, cmd, args=None, log_data=True):
        """
        Sends the given command, with optional arguments.

        :param str cmd: the command (eg. GETPIN, SETKEYINFO, etc.)
        :param iterable[str] args: command arguments.
        :param bool log_data: if False, the data in response won't be logged.
        :return:
        """
        if args is None:
            args = []

        command = cmd + ' ' + ' '.join(self._encode(x) for x in args) + '\n'

        self._log_output(command[:-1])

        self.output.write(command)
        self.output.flush()

        responses = list(self._responses(log_data=log_data))
        if responses[-1][0] == 'ERR':
            raise AssuanError(responses[-1][1])

        data = ''.join(r[1] for r in responses[:-1] if r[0] == 'D')
        return data, responses[-1]

    def bye(self):
        try:
            self.send('BYE')
        except (AssuanError, AssuanProtocolError) as e:
            log.error('Error while taking leave of %s: %s', self.name, str(e), exc_info=e)


@contextlib.contextmanager
def gpg_agent(socket_path: str) -> Assuan:
    with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as s:
        s.connect(socket_path)
        with s.makefile(mode='rw', encoding='utf-8', newline='\n') as fd:
            agent = Assuan('gpg-agent', fd, fd)
            try:
                yield agent
            finally:
                agent.bye()


class PinEntry(Assuan):
    """
    Helper for using pinentry to ask passwords to user.
    """
    def __init__(self, name, reader, writer):
        super().__init__(name, reader, writer)

    def set_keyinfo(self, keygrip):
        self.send('SETKEYINFO ', args=[keygrip])

    def set_error(self, errmsg='Invalid PIN.'):
        self.send('SETERROR', args=[errmsg])

    def ask_pin(self, prompt='PIN:'):
        self.send('SETPROMPT', args=[prompt])
        data, _ = self.send('GETPIN', log_data=False)
        return data


@contextlib.contextmanager
def pinentry(executable='pinentry'):

    def iowrap(rawio):
        return io.TextIOWrapper(rawio, encoding='utf-8', newline='\n')

    with subprocess.Popen([], executable=executable, stdout=subprocess.PIPE, stdin=subprocess.PIPE) as process:
        with iowrap(process.stdin) as writer, iowrap(process.stdout) as reader:
            pin_entry = PinEntry('pinentry', reader, writer)
            try:
                yield pin_entry
            finally:
                pin_entry.bye()


def kill_scdaemon(socket_path: str=None) -> bool:
    """
    Sends the SCD RESTART and SCD BYE commands to the gpg agent via the given socket. This makes the scdaemon
    to quit, allowing to use a yubikey otherwise used with gpg.
    :param socket_path: control socket for scdaemon.
    :return: True if the calls succeed, False otherwise
    """
    if socket_path is None:
        socket_path = '/run/user/{uid}/gnupg/S.gpg-agent'.format(uid=os.getuid())

    if not os.path.exists(socket_path):
        log.info('Could not kill scdaemon as gpg-agent socket path %s was not found!', socket_path)
        return False

    log.info('Killing scdaemon using socket %s', socket_path)

    with gpg_agent(socket_path) as agent:
        try:
            agent.send('SCD RESTART')
            agent.send('SCD BYE')
        except AssuanError as e:
            log.error('Unable to kill scdaemon: 0x%x %s', e.code, e.message, exc_info=e)
            return False

    return True



# coding=utf-8

# Copyright (C) 2017, Alexandre Vaissière
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
# OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import os
from setuptools import setup, find_packages

import oathguardian

ICON_SIZES = (16, 32, 48, 64, 96, 128, 192, 256, 512)

data_files = []
data_files += [('share/icons/hicolor/{size}x{size}/apps'.format(size=s),
                ['data/icons/{size}/oathguardian.png'.format(size=s)])
               for s in ICON_SIZES]
data_files += [('share/icons/hicolor/scalable/apps', ['data/oathguardian.svg'])]
data_files += [('share/applications', ['data/oathguardian.desktop'])]

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name=oathguardian.NAME,
    version=oathguardian.VERSION,
    description=oathguardian.DESCRIPTION,
    license=oathguardian.LICENSE_NAME,
    url=oathguardian.URL,

    long_description=long_description,
    long_description_content_type='text/markdown',

    author='Alexandre Vaissière',
    author_email='avaiss@fmiw.org',

    python_requires='~=3.5',
    keywords='totp authenticator oath gtk one-time-password',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: X11 Applications :: GTK",
        "Environment :: X11 Applications :: Gnome",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: ISC License (ISCL)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Utilities",
    ],

    packages=find_packages(exclude=['tests']),
    package_dir={'': '.'},
    package_data={
        'oathguardian.gtkui': ['oathguardian.xml', 'oathguardian_menu.xml'],
    },
    include_package_data=True,

    data_files=data_files,

    install_requires=['pyzbar>=0.1.4', 'yubikey-manager>=0.6.0', 'secretstorage>=2.3.1'],

    setup_requires=['nose==1.3.7'],

    entry_points={
        'console_scripts': [
        ],
        'gui_scripts': [
            'oathguardian-gtk = oathguardian.gtkui.__main__:main',
        ]
    },
)



# coding: utf-8

# region License

# Copyright (c) 2017,2018 Alexandre Vaissière <avaiss@fmiw.org>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# endregion

"""
Utilities for computing HOTP and TOTP codes.
"""

from .formatters import (formatter_type, new_default_formatter, new_human_formatter, new_steam_formatter)
from .keys import (key_types, to_binary_key, b32_normalize, to_b32_key)
from .oath import (generate_hotp, generate_totp)
from .providers import Provider, OATHType, HashAlgo, Credential, Code
from .uri import parse_uri



# coding: utf-8

# region License

# Copyright (c) 2017,2018 Alexandre Vaissière <avaiss@fmiw.org>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# endregion

import enum
import pkgutil
import types

import gi
gi.require_version('Gtk', '3.0')

from gi.repository import GdkPixbuf
from gi.repository import Gtk


def load_uifile(filename: str, package: str=__name__) -> Gtk.Builder:
    """
    Load a glade file and return the corresponding widget tree.

    :param filename: path to file, relative to package
    :param package: package path
    :return: GTK widget tree
    :except FileNotFoundError: if the file is not found.
    """
    data = pkgutil.get_data(package, filename)
    if not data:
        raise FileNotFoundError("File {} not found in package {}".format(filename, package))
    widget_tree = Gtk.Builder()
    widget_tree.add_from_string(data.decode('utf-8'))
    return widget_tree


def load_image(filename: str, package: str=__name__) -> GdkPixbuf.Pixbuf:
    """
    Load an image file from package data, and return the pixbuf data.

    :param filename: path to image, relative to package
    :param package: package path
    :return: the image as Pixbuf object
    :except FileNotFoundError: if the image file is not found.
    """
    data = pkgutil.get_data(package, filename)
    if not data:
        raise FileNotFoundError("File {} not found in package {}".format(filename, package))

    loader = GdkPixbuf.PixbufLoader()
    loader.write(data)
    return loader.get_pixbuf()


class Columns(enum.IntEnum):
    """

    """
    def __new__(cls, name: str, data_type: type):
        value = len(cls.__members__)

        obj = int.__new__(cls, value)
        obj._value_ = value
        obj._column_name = name
        obj._data_type = data_type

        return obj

    @property
    def display_name(self):
        return self._column_name

    @property
    def data_type(self):
        return self._data_type

    @classmethod
    def column_types(cls):
        return (column.data_type for column in cls)



# coding: utf-8

# region License

# Copyright (c) 2018 Alexandre Vaissière <avaiss@fmiw.org>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# endregion

"""

"""

import base64
import hmac
import time
import logging

from .keys import (key_types, b32_normalize)


def generate_hotp(key: key_types, counter: int, digits: int=6, hash_algo: str='sha1') -> int:
    """
    Generates an HOTP code following RFC 4226. The returned value is an int, that has to be formatted.

    Indeed, some issuers are not expecting the HOTP code to be formatted as documented (eg. steam).

    :param key: the key, either as a b32 string or a byte array.
    :param counter: the counter.
    :param digits: required number of digits, generally 6 or 8
    :param hash_algo: hash algorithm to use. For HOTP code it is sha1, as RFC 4226 is pretty strict about it. However,
                      for TOTP codes, sha256 or sha512 may be used.
    :return: the HOTP code, as int.
    """
    if counter < 0:
        raise ValueError('Counter must be positive')

    log = logging.getLogger(__name__)

    log.debug('Generating HOTP for counter=%d with key=%s, digits=%d, algo=%s', counter, key, digits, hash_algo)

    if isinstance(key, base64.bytes_types):
        key_bytes = key
    else:
        key_bytes = base64.b32decode(b32_normalize(key))

    counter_bytes = counter.to_bytes(8, byteorder='big')

    # gets the HMAC-SHA1 value of counter_bytes by key
    hmac_result = hmac.new(key_bytes, counter_bytes, digestmod=hash_algo).digest()

    # directly copied from https://www.ietf.org/rfc/rfc4226.txt section 5.4
    offset = hmac_result[-1] & 0xf
    bin_code = ((hmac_result[offset] & 0x7f) << 24
                | (hmac_result[offset+1] & 0xff) << 16
                | (hmac_result[offset+2] & 0xff) << 8
                | (hmac_result[offset+3] & 0xff))

    return bin_code % (10 ** digits)


def generate_totp(key: key_types, timestamp: int=None, period: int=30,
                  digits: int=6, hash_algo: str='sha1', origin: int=0) -> int:
    """
    Generates an TOTP code following RFC 6238. In fact, this is just computing an HOTP code by using
    ``(timestamp - origin) // period`` as counter.

    :param key: the key, either formatted as a b32 string or in binary.
    :param timestamp: the timestamp in seconds for which to compute the token. If not provided, current time will be
                      taken.
    :param period: time step in seconds.
    :param digits: required number of digits, generally 6 or 8
    :param hash_algo: hash algorithm to use. For HOTP code it is sha1, as RFC 4226 is pretty strict about it. However,
                      for TOTP codes, sha256 or sha512 may be used.
    :param origin: Unix time (in seconds) to start counting steps. Default is 0 (the Epoch).
    :return: the HOTP code, as int.
    """
    if timestamp is None:
        timestamp = int(time.time())

    steps = (timestamp - origin) // period

    return generate_hotp(key, steps, digits=digits, hash_algo=hash_algo)



# coding=utf-8

# region License

# Copyright (c) 2017,2018 Alexandre Vaissière <avaiss@fmiw.org>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# endregion


import typing

from gi.repository import Gtk

from oathguardian.gtkui.datamodel import ProviderColumns
from . import actions
from oathguardian.gtkui import datamodel
from oathguardian.oath import OATHType, HashAlgo


class AddDialog(object):
    def __init__(self, main_window, widgets):
        self.og_model = main_window.app.model  # type: datamodel.OGModel

        self.add_dialog = widgets.get_object('add_dialog')
        self.add_dialog.set_transient_for(main_window.window)

        self.add_oath_type = widgets.get_object('add_oath_type')
        self.add_label = widgets.get_object('add_label')
        self.add_issuer = widgets.get_object('add_issuer')
        self.add_secret = widgets.get_object('add_secret')
        self.add_algorithm = widgets.get_object('add_algorithm')
        self.add_digits = widgets.get_object('add_digits')
        self.add_period = widgets.get_object('add_period')
        self.add_counter = widgets.get_object('add_counter')
        self.add_touch = widgets.get_object('add_touch')
        self.add_provider = widgets.get_object('add_storage')

        scan_button = widgets.get_object('add_scan_button')
        scan_button.connect("clicked", actions.ScanQRAction(main_window))

        oath_type_store = Gtk.ListStore(str, str, object)
        for item in OATHType:
            oath_type_store.append([item.value, item.name, item])
        self.add_oath_type.set_model(oath_type_store)
        self.add_oath_type.set_id_column(0)
        self.add_oath_type.set_entry_text_column(1)

        algorithm_store = Gtk.ListStore(str, str, object)
        for item in HashAlgo:
            algorithm_store.append([item.value, item.name, item])
        self.add_algorithm.set_model(algorithm_store)
        self.add_algorithm.set_id_column(0)
        self.add_algorithm.set_entry_text_column(1)

        self.add_provider.set_model(self.og_model.providers)
        self.add_provider.set_id_column(0)
        self.add_provider.set_entry_text_column(1)
        self.add_provider.set_active_id(self.og_model.default_provider_id)

    def run(self):
        return self.add_dialog.run()

    def hide(self):
        return self.add_dialog.hide()

    def fill(self, **kwargs):
        oath_type = kwargs.get('oath_type')

        self.add_oath_type.set_active_id(oath_type)
        self.add_label.set_text(kwargs.get('label', ''))
        self.add_issuer.set_text(kwargs.get('issuer', ''))
        self.add_secret.set_text(kwargs.get('secret', ''))
        self.add_algorithm.set_active_id(kwargs.get('algorithm', '').lower())
        self.add_digits.set_value(kwargs.get('digits', 0))

        # Those should depend on oath_type
        self.add_period.set_value(kwargs.get('period', 0))
        self.add_counter.set_value(kwargs.get('counter', 0))

        self.add_provider.set_active_id(self.og_model.default_provider_id)

        # For now, always set require_touch to true.
        self.add_touch.set_state(True)

    def validate(self):
        oath_type_iter = self.add_oath_type.get_active_iter()
        if not oath_type_iter:
            raise Exception('No valid auth type')

        oath_type = self.add_oath_type.get_model()[oath_type_iter][2]
        label = self.add_label.get_text()
        secret = self.add_secret.get_text()
        digits = self.add_digits.get_value_as_int()
        issuer = self.add_issuer.get_text()
        touch = self.add_touch.get_state()

        result = {
            'oath_type': oath_type,
            'label': label,
            'secret': secret,
            'digits': digits,
            'issuer': issuer,
            'require_touch': touch,
        }

        algo_iter = self.add_algorithm.get_active_iter()
        if algo_iter:
            result['algorithm'] = self.add_algorithm.get_model()[algo_iter][2]

        if oath_type == OATHType.TOTP:
            result['period'] = self.add_period.get_value_as_int()

        if oath_type == OATHType.HOTP:
            result['counter'] = self.add_counter.get_value_as_int()

        provider_iter = self.add_provider.get_active_iter()
        if not provider_iter:
            raise Exception('No provider selected!')
        provider = self.og_model.providers[provider_iter][ProviderColumns.PROVIDER]

        return result, provider


class PasswordDialog(object):
    DEFAULT_PROMPT = 'Please enter password'

    def __init__(self, widgets):
        self._dialog = widgets.get_object('password_dialog')
        self._dialog.set_default_response(Gtk.ResponseType.OK)

        self._label = widgets.get_object('password_label')
        self._password = widgets.get_object('password_entry')

        self._password.connect('icon-press', PasswordDialog._show_password)
        self._password.connect('activate', self._activate)

    def ask_password(self, prompt: str=None) -> typing.Union[str, None]:
        """
        :return: password entered by user, or None if user cancelled the request.
        """
        self._label.set_text(prompt or PasswordDialog.DEFAULT_PROMPT)

        try:
            while True:
                resp = self._dialog.run()
                if resp != Gtk.ResponseType.OK:
                    return None

                text = self._password.get_text()
                if text:
                    return text
        finally:
            self._reset()
            self._dialog.hide()

    def _reset(self):
        self._password.set_text('')

    def _activate(self, *args):
        self._dialog.response(Gtk.ResponseType.OK)

    @classmethod
    def _show_password(cls, entry, pos, *args, **kwargs):
        if pos == Gtk.EntryIconPosition.SECONDARY:
            entry.set_visibility(not entry.get_visibility())
